#include "TeamDecision.h"
#include "GameView.h"
#include "BallView.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "ActuatorInterface.h"
#include "ViewManager.h"
#include "view_constants.h"
#include "InitCommand.h"
#include "MoveCommand.h"
#include "SayCommand.h"
#include "DashCommand.h"
#include "KickCommand.h"
#include "ByeCommand.h"
#include "TurnCommand.h"
#include "TurnNeckCommand.h"
#include "rs_assert.h"
#include "rs_log.h"
#include <iostream.h>
#include <fstream.h>
#include "/home/jonatan/robocup/v2/evolve/genTypes.h"

RS_BEGIN_NAMESPACE

TeamDecision::TeamDecision(ActuatorInterface* const actuator_interface)
  : Decision(actuator_interface),
    viewManager(ViewManager::instance()),
    worldModelInterface(WorldModelInterface::instance()),
    facts(worldModelInterface->getWorldFacts()),
    gameView(NULL),
    ballView(NULL),
    agentView(NULL),
    playerView(NULL) {
  getView(gameView, GAME_VIEW_ID);
  getView(ballView, BALL_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
  dummyCommand = new TurnCommand(10);
  waitCommand = new TurnCommand(0);
}

TeamDecision::~TeamDecision() {
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
  delete dummyCommand;
  delete waitCommand;
}

/*****************
* Utilites
*****************/

int hex2dec(char c) {
  if (c>='0' && c<='9')
    return c-'0';
  return 10+c-'a';
}

char dec2hex(int i) {
  char h[]={'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
  return h[i];
}

/*****************
* Special Variables
*****************/
int  _time_weareattacking=0;
int  _time_opponentsareattacking=0;
int  _time_wearespreadout=0;
int  _time_teammatesarefree=0;
bool _v_weareattacking;
bool _v_opponentsareattacking;
bool _v_wearespreadout;
bool _v_teammatesarefree;

/*****************
* Gentic Stuff...
*****************/

void TeamDecision::initGen() {
  int i, blocksize;
  TreeType treeType;
  //Determine agent's invidual number in current population 
  const std::string teamname = gameView->getOurTeamName();
  invnumber = 16*hex2dec(teamname[playernumber*2-2]);
  invnumber +=   hex2dec(teamname[playernumber*2-1]);
  cout<<"Invnumber "<<invnumber<<endl;
  // Prepare Fitness evaluation
  strcpy(fitness_filename,"l0.fit");
  fitness_filename[0]=side;
  fitness_filename[1]=dec2hex(playernumber);
  cout<<"Fitness will be saved to "<<fitness_filename<<endl;
  fitness.inv          =invnumber;
  fitness.ourScore     =0;
  fitness.theirScore   =0;
  fitness.scoringShoots=0;
  fitness.kicks        =0;
  fitness.passes       =0;
  lastKickTime         =0;
  //read decision trees from gen.dat
  ifstream f;
  f.open("gen.dat",ios::in|ios::binary);
  f.read((char *)&i,sizeof(int)); 
  cout<<"Population Size = "<<i<<endl;
  f.seekg((long)f.tellg()+sizeof(int)+sizeof(bool)+i*(sizeof(float)+2*sizeof(int)),ios::beg);
  for (i=0; i<invnumber; ++i) {
    f.read((char *)&blocksize,sizeof(int));
    f.seekg((long)f.tellg()+sizeof(TreeType)+blocksize*sizeof(dTree));
    f.read((char *)&blocksize,sizeof(int));
    f.seekg((long)f.tellg()+sizeof(TreeType)+blocksize*sizeof(dTree));
  }
  //read move tree
  f.read((char *)&blocksize,sizeof(int));
  f.read((char *)&treeType,sizeof(TreeType));
  if (treeType!=MOVE)
    cout<<"ERROR: reading from gen.dat\n";
  dMove = new dTree[blocksize];
  for (i=0; i<blocksize; ++i)
    f.read((char *)&dMove[i],sizeof(dTree));
  //read kick tree
  f.read((char *)&blocksize,sizeof(int));
  f.read((char *)&treeType,sizeof(TreeType));
  if (treeType!=KICK)
    cout<<"ERROR: reading from gen.dat\n";
  dKick = new dTree[blocksize];
  for (i=0; i<blocksize; ++i)
    f.read((char *)&dKick[i],sizeof(dTree));
  f.close();
  //init predicates
  preds = new int[preds_stop+1];
  //init skills/actions
  currentlyPersistent = 0;
}


bool TeamDecision::pred(Predicates p) {
  //ballView->getPosition().isKnown() should always be TRUE
  if (!ballView->getPosition().isKnown()) {
    //cout<<"ball UNKNOWN\n";
    return false;
  }
  if (p < _EndCommonPredicates) {
    if (p == OpponentIsClose) {
      /*
	Opponent Is Close?
	1 parameters: range
      */
      const int range = 5;
      PointUE ap = agentView->getPosition();
      if (ap.isUnknown()) {
	return true;
      }
      PointUE po = (playerView->getOpponentClosestToAgent()).getPosition();
      if (po.isKnown()) {
	if ((po.getPoint()).getDistanceTo(ap.getPoint()) < range) {
	  return true;
	}
      }
      return false;
    }
    /*
      I Am Near Opponent Goal
    */
    else if (p == IAmNearGoal) {
      const Float32 range = 25.0;
      Point ap = agentView->getPosition().getPoint();
      return ap.getDistanceTo(facts->getTheirGoal().getPosition()) < range;
    }
    /*
      I Am Near Our Goal
    */
    else if (p == IAmNearOurGoal) {
      const Float32 range = 25.0;
      Point ap = agentView->getPosition().getPoint();
      return ap.getDistanceTo(facts->getOurGoal().getPosition()) < range;
    }
    /*
      I Am Alone
    */
    else if (p == IAmAlone) {
      const Float32 range = 15.0;
      PointUE pp = playerView->getPlayerClosestToAgent().getPosition();
      Point ap = agentView->getPosition().getPoint();
      if (pp.isUnknown())
	return true;
      return ap.getDistanceTo(pp.getPoint()) > range;
    }
    /*
      We Are Wining?
    */
    else if (p == WeAreWining) {
      return ( gameView->getOurScore()-1>gameView->getTheirScore() );
    }
    /*
      We Are Attacking
    */
    else if (p == WeAreAttacking) {
      int gtime = (int)(gameView->getCurrentTime()).getT();
      if (gtime < _time_weareattacking)
	return _v_weareattacking;
      _time_weareattacking = gtime+9;
      
      PVector pl=playerView->getTeammates();
      PointUE q = agentView->getPosition();
      int cnt=0;
      if (q.isKnown()) {
	if (q.getPoint().getX() > 30.0) {
	  _v_weareattacking = true;
	  return true;
	}
	else if (q.getPoint().getX() > 15.0)
	  cnt++;
      }
      for (PVector::iterator i = pl.begin(); i<pl.end(); ++i) {
	q = (*i).getPosition();
	if (q.isKnown()) {
	  if (q.getPoint().getX() > 15.0) {
	    cnt++;
	    if (cnt == 2) {
	      _v_weareattacking = true;
	      return true;
	    }
	  }
	}
      }
      _v_weareattacking = false;
      return false;
    }
    /*
      Opponents Are Attacking
    */
    else if (p == OpponentsAreAttacking) {
      int gtime = (int)(gameView->getCurrentTime()).getT();
      if (gtime < _time_opponentsareattacking)
	return _v_opponentsareattacking;
      _time_opponentsareattacking = gtime+8;

      PVector pl=playerView->getOpponents();
      PointUE q;
      int cnt=0;
      for (PVector::iterator i = pl.begin(); i<pl.end(); ++i) {
	q = (*i).getPosition();
	if (q.isKnown()) {
	  if (q.getPoint().getX() < -15.0) {
	    cnt++;
	    if (cnt == 2) {
	      _v_opponentsareattacking = true;
	      return true;
	    }
	  }
	}
      }
      _v_opponentsareattacking = false;
      return false;
    }
    /*
      Teammates Are Free
      (box size: 10)
    */
    else if (p == TeammatesAreFree) {
      int gtime = (int)(gameView->getCurrentTime()).getT();
      if (gtime < _time_teammatesarefree)
	return _v_teammatesarefree;
      _time_teammatesarefree = gtime+5;

      PVector pl=playerView->getTeammates();
      PVector ol=playerView->getOpponents();
      if (ol.size()==0) {
	_v_teammatesarefree = true;
	return true;
      }
      if (pl.size()==0) {
	_v_teammatesarefree = false;
	return false;
      }
      PVector::iterator i,j;
      PointUE q;
      int cnt=0;
      float x,y;
      for (i = pl.begin(); i<pl.end(); ++i) {
	q = (*i).getPosition();
	if (q.isKnown()) {
	  x = q.getPoint().getX();
	  y = q.getPoint().getY();
	  for (j = ol.begin(); j<ol.end(); ++j) {
	    q = (*j).getPosition();
	    if (q.isKnown()) {
	      if (q.getPoint().getX()-x>10 || q.getPoint().getX()-x<-10) {
		continue;
	      }
	      if (q.getPoint().getY()-y>10 || q.getPoint().getY()-y<-10) {
		continue;
	      }
	      break;
	    }
	  }
	  if (j == ol.end()) {
	    _v_teammatesarefree = true;
	    return true;
	  }
	}
      }
      _v_teammatesarefree = false;
      return false;
    }
    /*
      We Are Spread Out
    */
    else if (p == WeAreSpreadOut) {
      int gtime = (int)(gameView->getCurrentTime()).getT();
      if (gtime < _time_wearespreadout)
	return _v_wearespreadout;
      _time_wearespreadout = gtime+7;

      PVector pl=playerView->getTeammates();
      PointUE q = agentView->getPosition();
      if (pl.size()==0) {
	_v_wearespreadout = true;
	return true;
      }
      if (pl.size()==1) {
	if (q.isUnknown() || (*pl.begin()).getPosition().isUnknown()) {
	  _v_wearespreadout = true;
	  return true;
	}
	if (q.getPoint().getDistanceTo((*pl.begin()).getPosition().getPoint())>15.0) {
	  _v_wearespreadout = true;
	  return true;
	}
	_v_wearespreadout = false;
	return false;
      }
      int n=0;
      float x=0.0,x2=0.0,y=0.0,y2=0.0,t;
      /*
      if (q.isKnown()) {
	t   = q.getPoint().getX();
	x  += t;
	x2 += t*t;
	t   = q.getPoint().getY();
	y  += t;
	y2 += t*t;
	n++;
      }
      else
	return true;
      */
      for (PVector::iterator i = pl.begin(); i<pl.end(); ++i) {
	q = (*i).getPosition();
	if (q.isKnown()) {
	  t   = q.getPoint().getX();
	  x  += t;
	  x2 += t*t;
	  t   = q.getPoint().getY();
	  y  += t;
	  y2 += t*t;
	  n++;
	}
      }
      if (n>0) {
	if (x2-x*x/n+y2-y*y/n>n*200) {
	  _v_wearespreadout = true;
	  return true;
	}
	_v_wearespreadout = false;
	return false;
      }
      _v_wearespreadout = true;
      return true;
    }
  }
  else {
    if (p == IAmClosestToBallOnOurTeam) {
      /*
	I Am Closest To Ball On Our Team?
      */
      if (!agentView->getPosition().isKnown())
	return false;
      PointUE player_point = playerView->getTeammateClosestToBall().getPosition();
      if (player_point.isKnown()) {
	Point ballPoint = ballView->getPosition().getPoint();
	Float32 pd = player_point.getDistanceTo(ballPoint).getValue();
	Float32 ad = ballView->getRelativeDistance().getValue();
	if (pd > ad)
	  return true;
      }
      else
	return true;
      return false;
    }
    else if (p == WeAreClosestToBall) {
      /*
	We Are Closest To Ball?
	1 parameter: range (set to 10.0)
      */
      PointUE op = playerView->getOpponentClosestToBall().getPosition();
      if (op.isUnknown())
	return true;
      PointUE tp = playerView->getTeammateClosestToBall().getPosition();
      Point bp = ballView->getPosition().getPoint();
      Float32 td = ballView->getRelativeDistance().getValue();
      Float32 od;
      if (tp.isKnown()) {
	od = tp.getDistanceTo(bp).getValue();
	if (od < td)
	  td = od;
      }  
      od = op.getDistanceTo(bp).getValue();
      if (td < od && td < 10.0)
	return true;
      else
	return false;
    }
    /*
      Ball Is Close?
      1 parameter: range
    */
    else if (p == BallIsClose) {
      const Float32 range = 5.0;
      Point ap = agentView->getPosition().getPoint();
      Point bp = ballView->getPosition().getPoint();
      return ap.getDistanceTo(bp) < range;
    }
    /*
      Ball Is Far Away?
      1 parameter: range
    */
    else if (p == BallIsFarAway) {
      const Float32 range = 35.0;
      Point ap = agentView->getPosition().getPoint();
      Point bp = ballView->getPosition().getPoint();
      return ap.getDistanceTo(bp) > range;
    }
    /*
      Ball Is Near Opponent Goal
    */
    else if (p == BallIsNearGoal) {
      const Float32 range = 32.0;
      Point bp = ballView->getPosition().getPoint();
      return bp.getDistanceTo(facts->getTheirGoal().getPosition()) < range;
    }
    /*
      Ball Is Near Our Goal
    */
    else if (p == BallIsNearOurGoal) {
      const Float32 range = 32.0;
      Point bp = ballView->getPosition().getPoint();
      return bp.getDistanceTo(facts->getOurGoal().getPosition()) < range;
    }
    /*
      I Am Surrounded
    */
    /*
      else if (p == IAmSurrounded) {
      PVector pl=playerView->getOpponents();    
      Point ap = agentView->getPosition().getPoint();
      Vector vc;
      AngleDeg ang;
      bool zones[]={0,0,0,0};
      for (PVector::iterator i=pl.begin(); i!=pl.end(); ++i) {
      if (!(*i).getPosition().isKnown()) 
      continue;
      vc = ap.getVectorTo((*i).getPosition().getPoint());
      ang = agentView->getBodyDirection().getDeg()-vc.getAngle();
      if (ang>-90 && ang<90)
      if (ang<-45)
      zones[0]=1;
      else if (ang < 0)
      zones[1]=1;
      else if (ang < 45)
      zones[2]=1;
      else
      zones[3]=1;
      }
      return (zones[0]+zones[1]+zones[2]+zones[3]) > 2;
      }
    */
    /*
      UNKNOWN 
    */
    else {
      //cout<<"Unknown Predicate\n";
      return true;
    }
  }
}

inline void TeamDecision::setCurrentAction(Actions a) {
  currentAction = a;
}

void TeamDecision::preformAction() {
  // TODO! Make sure agent's position + dir and the same for ball is KNOWN
  if (!agentView->getPosition().isKnown() || 
      !ballView->getPosition().isKnown()) {
    //cout<<"Player: "<<playernumber<<", Ball or Agent UNKNOWN POSITION\n";
    decideNoAction();
    return;
  } 
  /*
    Dash To Ball
  */
  if (currentAction == DashToBall) {
    Point pBall;
    if (!ballView->getPosition(NEXT).isKnown()) 
      pBall = ballView->getPosition().getPoint();
    else
      pBall = ballView->getPosition(NEXT).getPoint();
    ServerCommand *cmd;
    AngleDeg turnAngle = (agentView->getPosition().getDirectionTo(pBall) -
			  agentView->getBodyDirection()).getAngle();
    if(abs(turnAngle) > AngleDeg(8.0)) {
      cmd = new TurnCommand(turnAngle);
      currentlyPersistent = 1;
    }
    else 
      cmd = new DashCommand(60); 
    actuatorInterface->addToPlan(cmd);
    delete cmd;  
  }
  /*
    Dash To Enemy Goal 
  */
  else if (currentAction == DashToGoal) {
    Point pGoal = Point(35,playernumber%2==0?20:-20);
    ServerCommand *cmd;
    AngleDeg turnAngle = (agentView->getPosition().getDirectionTo(pGoal) -
			  agentView->getBodyDirection()).getAngle();
    if(abs(turnAngle) > AngleDeg(5.0)) {
      cmd = new TurnCommand(turnAngle);
      currentlyPersistent = 1;
    }
    else 
      cmd = new DashCommand(60); 
    actuatorInterface->addToPlan(cmd);
    delete cmd;  
  }
  /*
    Block Home Goal
    TODO: Calculate exact point
  */
  else if (currentAction == BlockGoal) {
    Point pBall;
    Point pGoal = facts->getOurGoal().getPosition();
    Point ap = agentView->getPosition().getPoint();
    if (!ballView->getPosition(NEXT).isKnown()) 
      pBall = ballView->getPosition().getPoint();
    else
      pBall = ballView->getPosition(NEXT).getPoint();
    Point target;
    target.setX(0.7*pBall.getX()+0.3*pGoal.getX());
    target.setY(0.7*pBall.getY()+0.3*pGoal.getY());
    //target /= 2.0;
    //cout<<pGoal<<"  -  "<<pBall<<"  -  "<<target<<endl;
    ServerCommand *cmd;
    AngleDeg turnAngle;
    if (ap.getDistanceTo(pBall) < 7) {
      turnAngle = (agentView->getPosition().getDirectionTo(pBall)).getAngle();
      turnAngle-= (agentView->getBodyDirection()).getAngle();
    }
    else {
      turnAngle = ap.getDirectionTo(target);
      turnAngle-= (agentView->getBodyDirection()).getAngle();
    }
    if(abs(turnAngle) > AngleDeg(8.0)) {
      cmd = new TurnCommand(turnAngle);
      currentlyPersistent = -1;
      persistentCmd = new DashCommand(60);
    }
    else
      cmd = new DashCommand(60);
    actuatorInterface->addToPlan(cmd);
    delete cmd;  
  }
  /*
    Intercept Ball
  */
  else if (currentAction == InterceptBall) {
    if (receivePass.applicable()) {
      static Plan plan = receivePass.generatePlan();
      actuatorInterface->addToPlan(plan.getNextStep());
    }
    else {
      setCurrentAction(DashToBall);
      preformAction();
      setCurrentAction(InterceptBall);
    }
  }
  /*
    Spread Team
    1 parameter: radius
  */
  else if (currentAction == SpreadTeam) {
    const Float32 radius = 15.0;
    PVector pl=playerView->getTeammates();    
    Point ap = agentView->getPosition().getPoint();
    Vector vc(0,0);
    AngleDeg turnAngle(90);
    Point pp;
    for (PVector::iterator i=pl.begin(); i!=pl.end(); ++i) {
      if (!(*i).getPosition().isKnown()) 
	continue;
      pp = (*i).getPosition().getPoint();
      if (ap.getDistanceTo(pp) < radius)
	vc += pp.getVectorTo(ap);
    }
    if (vc.getLength() > 1.0) {
      turnAngle = vc.getAngle().getDeg();
      turnAngle-= agentView->getBodyDirection().getAngle().getDeg();
    }
    currentlyPersistent = -5;
    persistentCmd = new DashCommand(60);
    ServerCommand *cmd = new TurnCommand(turnAngle);
    actuatorInterface->addToPlan(cmd);
    delete cmd;
  }
  /*
    -----------------------------
    ------- Kick Skills ---------
    -----------------------------
  */
  /*
    Pass Skills
    1 parameter: kickPowerWeigth
  */
  else if (currentAction >= PassPlayer1 && currentAction <= PassPlayerFarAway){
    const float kickPowerWeigth = 5.0;
    Point target;
    Float32 distance;
    VectorUE targetSpeed;
    Point ap = agentView->getPosition().getPoint();
    // passPlayer1
    if (currentAction == PassPlayer1) {
      PointUE pp = playerView->getTeammateClosestToAgent().getPosition();
      if (pp.isKnown())
	target = pp.getPoint();
      else {
	//setCurrentAction(KickGoal);
	//preformAction();
	decideNoAction();
	return;
      }
      targetSpeed = playerView->getTeammateClosestToAgent().getSpeedVector();
    }
    // passPlayer2
    else if (currentAction == PassPlayer2) {
      PVector pl=playerView->getTeammates();    
      const Float32 maxDis = 10000.0;
      Float32 dis, dis1=maxDis, dis2=maxDis;
      VectorUE sv1;
      Point pp, pp1, pp2;
      for (PVector::iterator i=pl.begin(); i!=pl.end(); ++i) {
	if (!(*i).getPosition().isKnown()) 
	  continue;
	pp = (*i).getPosition().getPoint();
	dis = ap.getDistanceTo(pp);
	if (dis < dis1) {
	  dis2 = dis1;
	  pp2 = pp1;
	  targetSpeed = sv1;
	  dis1 = dis;
	  pp1 = pp;
	  sv1 = (*i).getSpeedVector();
	}
	else if (dis < dis2) {
	  dis2 = dis;
	  pp2 = pp;
	  targetSpeed = (*i).getSpeedVector();
	}
      }
      if (dis2 == maxDis) {
	if (dis1 == maxDis) {
	  //setCurrentAction(KickGoal);
	  //preformAction();
	  decideNoAction();
	  return ;
	}
	pp2 = pp1;
	targetSpeed = sv1;
      }
      target = pp2;
    }
    // passToPlayerFarAway
    else { //if (currentAction==PassToPlayerFarAway) {
      PVector pl=playerView->getTeammates();    
      const Float32 maxDis = 10000.0;
      Float32 dis, dis1=maxDis;
      Point pp, pp1;
      for (PVector::iterator i=pl.begin(); i!=pl.end(); ++i) {
	if (!(*i).getPosition().isKnown()) 
	  continue;
	pp = (*i).getPosition().getPoint();
	dis = pp.getDistanceTo(facts->getTheirGoal().getPosition());
	if (dis < dis1) {
	  dis1 = dis;
	  pp1 = pp;
	  targetSpeed = (*i).getSpeedVector();
	} 
      }
      if (dis == maxDis) {
	//setCurrentAction(KickGoal);
	//preformAction();
	decideNoAction();
	return ;
      }
      target = pp1;
    }
    if (!targetSpeed.isKnown()) {
      targetSpeed = Vector(0,0);
    }
    Vector kickVector = ap.getVectorTo(target) + targetSpeed.getVector();
    kickVector = kickVector.rotate(-(agentView->getBodyDirection().getAngle()));
    char msg[] = "l00000";
    if (side == 'r')
      msg[0]='r';
    int cx = (int)target.getX()+54;
    int cy = (int)target.getY()+32;
    ServerCommand *cmd;
    cmd = new KickCommand(kickPowerWeigth*kickVector.getLength(),kickVector.getAngle());
    actuatorInterface->addToPlan(cmd);
    delete cmd;
    if (cx>=0 && cy>=0) {
      if (cx>=100) 
	msg[1]='1';
      if (cx>=10)
	msg[2]='0'+(cx/10)%10;
      msg[3]='0'+cx%10;
      if (cy>=10)
	msg[4]='0'+(cy/10)%10;
      msg[5]='0'+cy%10;
      cmd = new SayCommand(msg);
      actuatorInterface->addToPlan(cmd);
      delete cmd;
    }
    fitness.passes++;
  }
  /*
    Beckham Pass
  */
  else if (currentAction==PassBeckham) {
    if (beckhamPass.applicable()) {
      fitness.passes++;
      static Plan plan = beckhamPass.generatePlan();
      actuatorInterface->addToPlan(plan.getNextStep());
    }
    else {
      //cout<<"NO BECKHAM\n";
      decideNoAction();
      //setCurrentAction(KickGoal);
      //preformAction();
    }
  }
  /*
    KickGoal
    1 parameter: kickPower
  */
  else if (currentAction==KickGoal) {
    const int kickPower = 60;
    AngleDegUE goal_dir = agentView->getPosition().getDirectionTo(facts->getTheirGoal().getPosition());
    ServerCommand* cmd;
    cmd = new KickCommand(kickPower,(goal_dir-agentView->getBodyDirection()).getAngle());
    actuatorInterface->addToPlan(cmd);
    delete cmd;
  }
  /*
    KickCenter
    1 parameter: kickPower
  */  /*
  else if (currentAction==KickGoal) {
    const int kickPower = 50;
    AngleDegUE goal_dir = agentView->getPosition().getDirectionTo(Point(0,0));
    ServerCommand* cmd;
    cmd = new KickCommand(kickPower,(goal_dir-agentView->getBodyDirection()).getAngle());
    actuatorInterface->addToPlan(cmd);
    delete cmd;
  } */
  /*
    Dribble To Goal
  */
  else if (currentAction == DribbleToGoal) {
    dribble.setDribblePoint(facts->getTheirGoal().getPosition());
    if (dribble.applicable()) {
      static Plan plan = dribble.generatePlan();
      actuatorInterface->addToPlan(plan.getNextStep());
    }
    else {
      setCurrentAction(KickGoal);
      preformAction();
      setCurrentAction(DribbleToGoal);
    }
  }
  /*
    Dribble Away
  */
  //else if (currentAction == DribbleAway) {
  //  
  //}
  else {
    decideNoAction();
  }
}

bool TeamDecision::persistent() {
  // MoveTowardsBall always false (no?!?)
  if (currentlyPersistent == 0)
    return false;
  if (currentlyPersistent > 0) {
    currentlyPersistent--;
    preformAction();
    return true;
  }
  if (currentlyPersistent < 0) {
    actuatorInterface->addToPlan(persistentCmd);
    currentlyPersistent++;
    preformAction();
  }
  if (currentlyPersistent == 0) {
    delete persistentCmd;
  }
  return true;
}

void TeamDecision::genDecide(const TreeType treeType) {
  dTree *tree;
  int pos = 0;
  int i;
  NodeID id;
  for (i=preds_start; i<=preds_stop; ++i)
    preds[i] = -1;
  if (treeType == MOVE) {
    tree = dMove;
  }
  else {
    tree = dKick;
  }
  //cout<<"-----------\n";
  while (tree[pos].type == PREDICATE) {
    id = tree[pos].id;
    if (preds[(int)id.predicate]==-1) {
      preds[(int)id.predicate]=pred(id.predicate);
    }
    //cout<<predicatesStrings[(int)id.predicate]<<endl;
    if (preds[(int)id.predicate]) {
      pos = tree[pos].next1;
    }
    else {
      pos = tree[pos].next2;
    }
  }
  //cout<<"   "<<actionStrings[tree[pos].id.action]<<endl;
  currentAction = tree[pos].id.action;
}

void TeamDecision::exitGen() {
  // save fitness struct
  ofstream f;
  f.open(fitness_filename,ios::out|ios::binary|ios::trunc);
  f.write((char *)&fitness,sizeof(Fitness));
  f.close();
  cout<<"Fitness saved to "<<fitness_filename<<endl;
  // save position history
  ofstream g;
  fitness_filename[3]='p';
  fitness_filename[4]='o';
  fitness_filename[5]='s';
  g.open(fitness_filename,ios::out|ios::binary|ios::trunc);
  cout<<"Position history saved to "<<fitness_filename<<endl;
  int i;
  //int gtime = (int)(gameView->getCurrentTime()).getT();
  for (int t = 0; t<=(int)(gameView->getCurrentTime()).getT(); ++t) {
    ppos[t].cycle=t;
    g.write((char*)&ppos[t],sizeof(Position));
  }
  /*
  Position p;
  while (t<=gtime) {
    gtime = (int)(gameView->getCurrentTime()).getT();
    p.cycle = t;
    if (gameView->getPlayMode(gtime-t) != PM_PLAY_ON) {
      p.playon = false;
    }
    else {
      p.playon = true;
      if (agentView->getPosition(gtime-t).isUnknown()) {
	p.agentPos.isKnown = false;
      }
      else {
	p.agentPos.isKnown = true;
	p.agentPos.x = agentView->getPosition(gtime-t).getPoint().getX();
	p.agentPos.y = agentView->getPosition(gtime-t).getPoint().getY();
      }
      if (ballView->getPosition(gtime-t).isUnknown()) {
	p.ballPos.isKnown = false;
      }
      else {
	p.ballPos.isKnown = true;
	p.ballPos.x = ballView->getPosition(gtime-t).getPoint().getX();
	p.ballPos.y = ballView->getPosition(gtime-t).getPoint().getY();
      }
    }
    g.write((char*)&p,sizeof(Position));
    t++;
  }
  */
  g.close();
  // Free memory;
  delete dMove;
  delete dKick;
  delete preds;
}

/*****************
* Decide an action depending on current game mode (e.g. mode != PLAY_ON)
*****************/

inline void TeamDecision::decideNoAction() {
  actuatorInterface->addToPlan(dummyCommand);
}

inline void TeamDecision::decideWait() {
  actuatorInterface->addToPlan(waitCommand);
}

int decide_kickoff_mem=0;
void TeamDecision::decide_kickoff() {
  if (findBall.applicable()) {
    //if (!ballView->getLastObservationTime().isKnown())
    //  cout<<" (uk) ";
    //else
    //  cout<<" "<<ballView->getLastObservationTime()<<" ";
    //cout<<" finding ball\n";
    static Plan plan = findBall.generatePlan();
    actuatorInterface->addToPlan(plan.getNextStep()); 
    decide_kickoff_mem=0;
  }
  else if (ballView->getRelativeDistance() <= facts->SP_kickable_margin) {
    Point po = Point(-10,10);
    Point pa = Point(0,0);
    AngleDeg po_dir = pa.getDirectionTo(po);
    if (decide_kickoff_mem == 0) {
      //cout<<" turning\n";
      ServerCommand* cmd; 
      cmd = new TurnCommand(po_dir-agentView->getBodyDirection().getAngle());
      actuatorInterface->addToPlan(cmd);
      delete cmd;
      decide_kickoff_mem=1;
    }
    else if (decide_kickoff_mem<3) {
      //cout<<" waiting\n";
      decide_kickoff_mem++;
      decideWait();
    }
    else {
      //cout<<" kick\n";
      ServerCommand* cmd;
      cmd = new KickCommand(50,po_dir-agentView->getBodyDirection().getAngle());
      actuatorInterface->addToPlan(cmd);
      delete cmd;
      decide_kickoff_mem=0;
    }
  }
  else {
    //cout<<" moving...\n";
    //static Plan plan = gotoBall.generatePlan();
    //actuatorInterface->addToPlan(plan.getNextStep());
    setCurrentAction(DashToBall);
    preformAction();
    decide_kickoff_mem=0;
  }    
  //else {
  //  decideNoAction();
  //}
}

void TeamDecision::decide_freekick() {  
  if (findBall.applicable()) {
    static Plan plan = findBall.generatePlan();
    actuatorInterface->addToPlan(plan.getNextStep());     
  }
  else if (pred(IAmClosestToBallOnOurTeam)) {
    if (ballView->getRelativeDistance() <= facts->SP_kickable_margin) {
      setCurrentAction(PassPlayer1);
      preformAction();
      return ; 
      
      PlayerObject po = playerView->getTeammateClosestToAgent();
      ServerCommand* scmd = 0;
      if (po.getPosition().isKnown()) {
      	AngleDegUE po_dir = agentView->getPosition().getDirectionTo(po.getPosition());
        scmd = new KickCommand(50,(po_dir-agentView->getBodyDirection()).getAngle());
      }
      else {
        scmd = new TurnCommand(45);
      }
      actuatorInterface->addToPlan(scmd);
      delete scmd;
    }
    else if (ballView->getPosition(NEXT).isKnown()) {
      //static Plan plan = gotoBall.generatePlan();
      //actuatorInterface->addToPlan(plan.getNextStep());
      setCurrentAction(DashToBall);
      preformAction();
    }
    else
      decideWait();
  }
  else
    decideWait();
}

void TeamDecision::decide() {
  static Plan plan;
  
  if (findBall.applicable()) {
    //RS_LOG ( LA_DECISION, "FindBall" );
    plan = findBall.generatePlan();
    actuatorInterface->addToPlan(plan.getNextStep());     
  }
  else if (gotoBall.applicable()) {
    //RS_LOG ( LA_DECISION, "GotoBall" );
    plan = gotoBall.generatePlan();
    actuatorInterface->addToPlan(plan.getNextStep());     
  }
  else if ( gameView->isKickAllowed() && (ballView->getRelativeDistance() 
				       <= facts->SP_kickable_margin)) {
    setCurrentAction(KickGoal);
    preformAction();
    /*    AngleDegUE goal_dir = agentView->getPosition().getDirectionTo
	(facts->getTheirGoal().getPosition());
      ServerCommand* scmd = new KickCommand(50,(goal_dir - 
				 agentView->getBodyDirection()).getAngle());
      actuatorInterface->addToPlan(scmd);
      delete scmd;*/
    }
  else
    decideNoAction();
}

/*****************
* Event Handlers
*****************/

ServerCommand* TeamDecision::initAgent() {
  InitCommand* init_cmd = new InitCommand(facts->CP_team_name, 
					  facts->SP_version, 
					  facts->CP_goalie);
  return init_cmd;
}

bool TeamDecision::start() {
  RS_REQUIRE( initialized == true );
  return true;
}

void TeamDecision::onAuralCoachSensorData() {
  //const std::string msg = agentView->getLastCoachMessage();
  //cout<<"Coach msg: "<<msg.data()<<" - "<<msg<<endl;
  //if (msg[0]=='B' && msg[1]=='Y' && msg[2]=='E') {
  cout<<(gameView->getCurrentTime()).getT()<<". Coach message received... bye bye\n";
  exitGen();
  ServerCommand *cmd = new ByeCommand();
  actuatorInterface->addToPlan(cmd);
  delete cmd;
  //}
}

void TeamDecision::onAuralRefereeSensorData() {
 PlayMode mode = gameView->getPlayMode();
 switch(mode) {
 case PM_OUR_GOAL:
   //cout<<"WE SCORED\n";
   if (goals_size-1>fitness.ourScore)
     fitness.goals[fitness.ourScore]=lastKickTime;
   fitness.ourScore++;
   teleportToInitialPosition(0);
   break;
 case PM_THEIR_GOAL:
   fitness.theirScore++;
   teleportToInitialPosition(0);
   break;
 case PM_BEFORE_KICK_OFF:
 case PM_HALF_TIME:
   if ((gameView->getCurrentTime()).getT()==3000) {
     cout<<"Game is over... bye bye\n";
     exitGen();
     ServerCommand *cmd = new ByeCommand();
     actuatorInterface->addToPlan(cmd);
     delete cmd;
   }
   break;
 default:
   break;
 }
}

void TeamDecision::onAuralPlayerSensorData() {
  const std::string msg = agentView->getLastPlayerMessage();
  if (side == msg[0] && agentView->getPosition().isKnown()) {
    Point ap = agentView->getPosition().getPoint();
    int cx = (int)ap.getX()+54;
    int cy = (int)ap.getY()+32;
    cx -= 100*(msg[1]-'0');
    cx -=  10*(msg[2]-'0');
    cx -=     (msg[3]-'0');
    if (cx<-2 || cx>2)
      return;
    cy -=  10*(msg[4]-'0');
    cy -=     (msg[5]-'0');
    if (cy<-2 || cy>2)
      return;
    passtome=15;
  }
}

void TeamDecision::onAuralSelfSensorData() {
}

void TeamDecision::teleportToInitialPosition(int state) {
  Point p;
  switch (playernumber) {      
  case 1:  p = Point(-10,  0); break; // Attacker
  case 2:  p = Point(-10,-10); break;
  case 3:  p = Point(-10, 10); break;
  case 4:  p = Point(-25, 10); break; // Mid
  case 5:  p = Point(-25,-10); break;
  case 6:  p = Point(-20,  0); break;
  case 7:  p = Point(-35,-20); break; // Def
  case 8:  p = Point(-35, 20); break;
  case 9:  p = Point(-25,-20); break; // Not used...
  case 10: p = Point(-25, 20); break;
  case 11: p = Point(-35,  0); break;
  default: p = Point(-10,-10);
  }
  if (state==0) {
    MoveCommand* move_cmd = new MoveCommand(p);
    actuatorInterface->addToPlan(move_cmd);
    delete move_cmd;
  }
  else if (state==1 && agentView->getBodyDirection().isKnown()) {
    Point pb = Point(0,0);
    AngleDeg ball_dir = p.getDirectionTo(pb);
    //cout<<ball_dir<<endl<<(agentView->getBodyDirection()).getAngle();
    TurnCommand* cmd = new TurnCommand(ball_dir-(agentView->getBodyDirection()).getAngle());
    actuatorInterface->addToPlan(cmd);
    delete cmd;
  }
  else if (state==2) {
    PointUE bp = ballView->getPosition();
    PointUE ap = agentView->getPosition();
    if (bp.isUnknown() || ap.isUnknown())
      return;
    if ((agentView->getBodyDirection()).isUnknown())
      return;
    AngleDeg ball_dir = (ap.getPoint()).getDirectionTo(ap.getPoint());
    AngleDeg player_dir = (agentView->getBodyDirection()).getAngle();
    TurnCommand* cmd = new TurnCommand(ball_dir-player_dir);
    actuatorInterface->addToPlan(cmd);
    delete cmd;
  }
}

void TeamDecision::onInit() {
  playernumber = agentView->getUniformNumber().getValue();
  side = (gameView->getOurSide() == S_RIGHT)?'r':'l';
  cout<<side<<' '<<playernumber<<endl;
  passtome=0;
  initGen();
  lastMode=PM_UNKNOWN;
  
  teleportToInitialPosition(0);
}

void TeamDecision::onCommandWarning() {
  PlayMode mode = gameView->getPlayMode();
  int gtime = (int)(gameView->getCurrentTime()).getT();
  switch (mode) {
  case PM_PLAY_ON:
    // Test predicates
    /*
    if (playernumber==1 && side=='l') {
      if (pred(WeAreAttacking))
	cout<<gtime<<". We Are Attacking\n";
      if (pred(OpponentsAreAttacking))
	cout<<gtime<<". They Are Attacking\n";
      if (pred(TeammatesAreFree))
	cout<<gtime<<". Mates are free\n";
      if (pred(WeAreSpreadOut))
	cout<<gtime<<".We Are Spread Out\n";
    }
    */
    // Save Positions for evaulation
    ppos[gtime].playon = true;
    if (ballView->getPosition().isUnknown()) 
      ppos[gtime].ballPos.isKnown = false;
    else {
      ppos[gtime].ballPos.isKnown = true;
      ppos[gtime].ballPos.x = ballView->getPosition().getPoint().getX();
      ppos[gtime].ballPos.y = ballView->getPosition().getPoint().getY();
    }
    if (agentView->getPosition().isUnknown())
      ppos[gtime].agentPos.isKnown = false;
    else {
      ppos[gtime].agentPos.isKnown = true;
      ppos[gtime].agentPos.x = agentView->getPosition().getPoint().getX();
      ppos[gtime].agentPos.y = agentView->getPosition().getPoint().getY();
    }
    // ---
    if (!persistent()) {
      TimeUnknown ballT = ballView->getLastObservationTime();
      if ( agentView->getBodyDirection().isUnknown() || 
	   agentView->getNeckDirection().isUnknown() ) {
	//cout<<gtime<<". Player "<<playernumber<<", Agent Unknown\n";
	decideNoAction();
      }
      else if ( ballView->getPosition().isUnknown() || 
		ballT.isUnknown() || 
		(ballT.getT()+20<gtime && gtime%15<3) ) {
        /*
	if (playernumber==2)
	  if (ballT.isKnown())
	    cout << gtime << ".  where did the ball go?!?\n";
	  else
	    cout << gtime << ".  ball?!?\n";
	*/
	ServerCommand *cmd;
	if (gameView->getNextSeeTime() == gameView->getNextTime()) 
	  cmd = new TurnCommand(90);//(agentView->getNeckDirection() + 90 -
				//agentView->getBodyDirection()).getDeg());
	else
	  cmd = new TurnCommand(20);
	actuatorInterface->addToPlan(cmd);
	delete cmd;
      }
      else if (ballView->getRelativeDistance() <= facts->SP_kickable_margin) {
	lastKickTime=gtime;
	fitness.kicks++;
	passtome=0;
	if (score.applicable()) {
	  fitness.scoringShoots++;
	  //cout<<gtime<<". PLAYER "<<playernumber<<" TRIES TO SCORE\n";
	  static Plan plan = score.generatePlan();
	  actuatorInterface->addToPlan(plan.getNextStep());
	}
	else {
	  //cout << gtime << ". kicking ball...\n";
	  genDecide(KICK);
	  //if (playernumber!=2)
	  // setCurrentAction(PassPlayer1);
	  //else
	  //  setCurrentAction(KickGoal);
	  preformAction();
	}
      }
      else {
	if (passtome>0)
	  setCurrentAction(InterceptBall);
	else
	  genDecide(MOVE);
	//if (playernumber!=2)
	//setCurrentAction(DashToBall);
	//else
	//  setCurrentAction(DashToBall);
	preformAction();
      }
    }
    break;
  case PM_OUR_KICK_OFF:
     ppos[gtime].playon = false;
    if (lastMode != mode) {
      teleportToInitialPosition(1);
      if (playernumber == 1) {
	if (currentlyPersistent!=0) 
	  delete persistentCmd;
	persistentCmd = new TurnCommand(0);
	currentlyPersistent = -10;
      }
    }
    else if (playernumber == 1) {
      //cout<<gtime<<". my kickoff...";
      decide_kickoff();
    }
    else
      decideWait();
    break;
  case PM_THEIR_KICK_OFF:
     ppos[gtime].playon = false;
    if (lastMode != mode)
      teleportToInitialPosition(1);
    else
      decideWait();
    break;
  case PM_OUR_CORNER_KICK:
  case PM_OUR_KICK_IN:
  case PM_OUR_FREE_KICK:
    ppos[gtime].playon = false;
    decide_freekick();
    break;
  case PM_THEIR_CORNER_KICK:
  case PM_THEIR_KICK_IN:
  case PM_THEIR_FREE_KICK:
    ppos[gtime].playon = false;
    if (lastMode != mode)
      teleportToInitialPosition(2);
    else
      decideWait();
    break;
  default:
    ppos[gtime].playon = false;
    break;
    decideWait();
  } 
  lastMode=mode;
  passtome--;
}

//void TeamDecision::onNewCycle(const bool forced) {

//}

RS_END_NAMESPACE
