// DecisionTree1.cpp: implementation of the DecisionTree class.
//
//////////////////////////////////////////////////////////////////////

//#include "stdafx.h"
//#include "genTypes.h"
#include "DecisionTree1.h"
#include <stdlib.h>
#include <fstream.h>
#include <iostream.h>


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

DecisionTree::DecisionTree(Log *log, TreeType type, int seed)
{
  //log->addArea(logArea,"DecisionTree");
  this->log = log;
  treeType = type;
  srand(seed); 
  tree = new dTree[INITIAL_SIZE];  
  maxEntries = INITIAL_SIZE;
  reset();
  _test();
}

DecisionTree::DecisionTree(Log *log, TreeType type, int nEntries, int seed)
{
	treeType = type;
	srand(seed); 
	tree = new dTree[nEntries];
	maxEntries = nEntries;
	reset();
	_test();
}

DecisionTree::~DecisionTree()
{
	delete tree;
}

bool DecisionTree::createRandomSubTree(int depth, int minSize, int maxSize, int pos, enum NodeType type)
{
  //log->debug(1,"maxsize ",maxSize);
  //log->debug(1,"depth ",depth);
        if (maxSize > maxDepth)
		maxSize = maxDepth;
	if (type==UNKNOWN)
		type = randBool() ? ACTION : PREDICATE;
	// Check conditions for type
	if (type!=ACTION && depth>=maxSize)
		type=ACTION;
	else if (type==ACTION && depth<minSize)
		type=PREDICATE;
	// Set node parameters
	tree[pos].used  = true;
	tree[pos].level = depth;
	tree[pos].type  = type;
	tree[pos].next1 = -1;
	tree[pos].next2 = -1;
	// Build Tree
	tree[pos].id = randID(type);
	if (type==PREDICATE) {
		tree[pos].next1=this->getUnusedIndex();
		this->createRandomSubTree(depth+1,minSize,maxSize,tree[pos].next1,UNKNOWN);
		tree[pos].next2=this->getUnusedIndex();
		this->createRandomSubTree(depth+1,minSize,maxSize,tree[pos].next2,UNKNOWN);
	}
	return true;
}

int DecisionTree::getUnusedIndex()
{
	int i = 0;
	while (i<maxEntries) {
		if (!tree[i].used)
			return i;
		++i;
	}
	dTree *tmp = new dTree[maxEntries+INC_SIZE];
	for (i=0; i<maxEntries; ++i)
		tmp[i] = tree[i];
	for (i=0; i<maxEntries+INC_SIZE; ++i)
		tmp[i].used = false;
	delete tree;
	tree = tmp;
	i = maxEntries;
	maxEntries+=INC_SIZE;
	return i;
}

int DecisionTree::randInt(int min, int max)
{
	int v, diff, mv;
	diff = max-min;
	mv = ((0x8000/(diff+1))/(diff+1))-1;
	do {
		v = rand()%0x8000;
	} while (v>mv);
	return (v % (diff+1))+min;
}

bool DecisionTree::randBool()
{
	return (rand()%2)==0;
}

const char* DecisionTree::convertID2Str(NodeType type, NodeID id)
{
	if (type == PREDICATE)
		return predicatesStrings[(int)id.predicate];
	if (type == ACTION)
		return actionStrings[(int)id.action];
	return "NULL";
}

void DecisionTree::print()
{
	int param[4]={0,0,0,0};
	print(0,param,NULL);
	//cout<<param[0]<<' '<<param[1]<<' '<<param[2]<<' '<<param[3]<<' '<<endl;
}

void DecisionTree::print(ofstream *f)
{
	int param[4]={0,0,0,0};
	print(0,param,f);
	if (param[0]>0 && param[2]>0) {
		*f<<"# nodes:      "<<param[0]<<endl;
		*f<<"avg. depth:   "<<(float)(param[1]/param[0])<<endl;
		*f<<"# leaf nodes: "<<param[2]<<endl;
		*f<<"avg. # pred:  "<<(float)(param[3]/param[2])<<endl;
	}
}

void DecisionTree::print(int pos, int *param, ofstream *f)
{
	int i;
	if (!tree[pos].used)
		return;
	// collect data for stats
	param[0]++;
	param[1]+=tree[pos].level;
	if (tree[pos].type==ACTION) {
		param[2]++;
		param[3]+=tree[pos].level;
	}
	// print
	if (f==NULL) {
		for (i=1; i<tree[pos].level; ++i) cout<<"    ";
		if (tree[pos].type==ACTION)
			cout<<convertID2Str(tree[pos].type,tree[pos].id)<<endl;
		else {
			cout<<"IF "<<convertID2Str(tree[pos].type,tree[pos].id)<<endl;
			print(tree[pos].next1,param,f);
			for (i=1; i<tree[pos].level; ++i) cout<<"    ";
			cout<<"ELSE\n";
			print(tree[pos].next2,param,f);
		}
	}
	else {
		for (i=1; i<=tree[pos].level; ++i) *f<<"    ";
		if (tree[pos].type==ACTION)
			*f  <<convertID2Str(tree[pos].type,tree[pos].id)<<endl;
		else {
			*f<<"IF "<<convertID2Str(tree[pos].type,tree[pos].id)<<endl;
			print(tree[pos].next1,param,f);
			for (i=1; i<=tree[pos].level; ++i) *f<<"    ";
			*f<<"ELSE\n";
			print(tree[pos].next2,param,f);
		}	
	}
}

void DecisionTree::deleteSubTree(int pos)
{
	tree[pos].used=false;
	if (tree[pos].type==PREDICATE) { 
		this->deleteSubTree(tree[pos].next1);
		this->deleteSubTree(tree[pos].next2);
	}
}

void DecisionTree::save(ofstream *f)
{
	f->write((char *)&maxEntries,sizeof(int));
	f->write((char *)&treeType,sizeof(TreeType));
	int i;
	for (i=0; i<maxEntries; ++i)
		f->write((char *)&tree[i], sizeof(dTree));
}

// pos = position of block
// returns position of next block
void DecisionTree::load(ifstream *f) 
{
	int i, blocksize;
	f->read((char *)&blocksize,sizeof(int));
	f->read((char *)&treeType,sizeof(TreeType));
	if (blocksize > maxEntries) {
		delete tree;
		tree = new dTree[blocksize];
		maxEntries = blocksize;
	}
	else
		reset();
	for (i=0; i<blocksize; ++i)
		f->read((char *)&tree[i],sizeof(dTree));
}

void DecisionTree::reset()
{
	//log->debug(logArea,"reset()");
	int i;
	maxDepth = MAX_TREE_SIZE;
	for (i=0; i<maxEntries; ++i)
		tree[i].used = false;
}

NodeID DecisionTree::randID(NodeType type)
{
	NodeID id;
	int p1,p2,p3,p4,i;
	bool v = (type==ACTION);
	p1 =     (int)(v ? _StartCommonActions: _StartCommonPredicates) +1;
	p2 =     (int)(v ? _EndCommonActions  : _EndCommonPredicates)   -1;
	if (this->treeType==KICK) {
		p3 = (int)(v ? _StartKickActions  : _StartKickPredicates)   +1;
		p4 = (int)(v ? _EndKickActions    : _EndKickPredicates)     -1;
	}
	else {
		p3 = (int)(v ? _StartMoveActions  : _StartMovePredicates)   +1;
		p4 = (int)(v ? _EndMoveActions    : _EndMovePredicates)     -1;
	}
	i = randInt(p1,p2+p4-p3+1);
	if (i>p2)
		i += p3-p2-1;
	if (type==ACTION)
		id.action = (Actions)i;
	else
		id.predicate = (Predicates)i;
	return id;
}

DecisionTree* DecisionTree::clone()
{
	DecisionTree *d = new DecisionTree(log, treeType, rand());
	if (d->maxEntries != this->maxEntries) {
		delete d->tree;
		d->tree = new dTree[this->maxEntries];
		d->maxEntries = this->maxEntries;
	}
	for (int i=0; i<this->maxEntries; ++i)
		d->tree[i] = this->tree[i];
	return d;
}

DecisionTree** DecisionTree::crossover(DecisionTree* t)
{
	DecisionTree *d[2], *p[2];
	int p1, p2;
	d[0] = this->clone();
	d[1] = t->clone();
	p1 = randInt(1,d[0]->nNodes());
	p2 = randInt(1,d[1]->nNodes());
	p1 = d[0]->getPos(p1);
	p2 = d[1]->getPos(p2);
	p[0] = d[0]->cloneSubTree(p1);
	p[1] = d[1]->cloneSubTree(p2);
//	delete d[0];
//	delete d[1];
//	d[0] = p[0];
//	d[1] = p[1];
	if (d[0]->tree[p1].level+p[1]->depth(0)-1>MAX_TREE_SIZE)
		p[1]->reduce();
	if (d[1]->tree[p2].level+p[0]->depth(0)-1>MAX_TREE_SIZE)
		p[0]->reduce(); 
	d[0]->insert(p[1],p1);
	d[1]->insert(p[0],p2);
	return d;
}

int DecisionTree::nNodes()
{
	int i, n=0;
	for (i=0; i<maxEntries; ++i)
		if (tree[i].used)
			++n;
	return n;
}

int DecisionTree::getPos(int no)
{
	int i, n=0;
	for (i=0; i<maxEntries; ++i)
		if (tree[i].used) {
			++n;
			if (n==no)
				return i;
		}
	log->error(logArea,"getPos(...) no = ",no);
	return 0;
}

DecisionTree* DecisionTree::cloneSubTree(int pos)
{
	bool *used = new bool[this->maxEntries];
	int *index = new int[this->maxEntries];
	int i, n=0;
	for (i=0; i<this->maxEntries; ++i)
		used[i]=false;
	_cloneSubTree(pos,used);
	for (i=0; i<this->maxEntries; ++i)
		if (used[i]) {
			index[i]=n;
			++n;
		}
	DecisionTree *d = new DecisionTree(log,this->treeType,n,rand());
	for (i=0; i<this->maxEntries; ++i)
		if (used[i]) {
			d->tree[index[i]] = this->tree[i];
			d->tree[index[i]].level -= this->tree[pos].level-1;
			if (d->tree[index[i]].type == PREDICATE) {
				d->tree[index[i]].next1 = index[d->tree[index[i]].next1];
				d->tree[index[i]].next2 = index[d->tree[index[i]].next2];
			}
		}
	delete index;
	delete used;
	return d;
}

void DecisionTree::_cloneSubTree(int pos, bool *used)
{
 	if (!tree[pos].used)
		log->error(logArea,"_cloneSubTree(...)");
	used[pos]=true;
	if (tree[pos].type == PREDICATE) {
		_cloneSubTree(tree[pos].next1,used);
		_cloneSubTree(tree[pos].next2,used);
	}
}


void DecisionTree::insert(DecisionTree *t, int pos)
{
//	cout<<"\n-----------------------------------------------------\n";
	if (!tree[pos].used)
		log->error(logArea,"insert(...)");
	bool *used = new bool[t->maxEntries];
	int *index = new int[t->maxEntries];
	int i,j=0;
	int l = this->tree[pos].level;
	bool first=true;
	for (i=0; i<t->maxEntries; ++i)
		used[i]=false;
	t->_cloneSubTree(0,used);
//	cout<<"positions used: ";
//	for (i=0; i<maxEntries; ++i)
//		if (tree[i].used)
//			cout << i << ' ';
//	cout<<endl;
//	print();
//	cout<<"#nodes = "<<nNodes()<<endl;
	this->deleteSubTree(pos);
//	cout<<"#nodes after deletion = "<<nNodes()<<endl;
	for (i=0; i<t->maxEntries; ++i)
		if (used[i]) {
			j++;
			if (first) {
				index[0]=pos;
				first=false;
			}
			else
				index[i]=this->getUnusedIndex();
			if (this->tree[index[i]].used)
				log->error(0,"DecisionTree::insert - error in DecisionTree::getUnusedIndex()");
			this->tree[index[i]].used=true;
		}
//	cout<<"#nodes after allocation = "<<nNodes()<<endl;
	if (j==0)
		log->error(0,"DecisionTree::insert(...) - Tried to insert an empty tree");
//	t->print();
//	print();
//	cout<<"t->max...="<<t->maxEntries<<endl;
//	cout<<"index[0]="<<index[0]<<endl;
//	cout<<(int)ACTION<<endl<<(int)t->tree[0].type<<endl;
//	cout.flush();
//	j = t->depth(0);
	for (i=0; i<t->maxEntries; ++i)
		if (used[i]) {
//			cout<<"Insertion at node #"<<index[i]<<" from t's node #"<<i<<endl;
			this->tree[index[i]] = t->tree[i];
			this->tree[index[i]].level += l-1;
			if (this->tree[index[i]].type == PREDICATE) {
				this->tree[index[i]].next1 = index[this->tree[index[i]].next1];
				this->tree[index[i]].next2 = index[this->tree[index[i]].next2];
				if (!tree[tree[index[i]].next1].used || !tree[tree[index[i]].next2].used)
					log->error(1,"DecisionTree::insert(...)");
			}
		}
//	cout<<"#nodes after insertions = "<<nNodes()<<endl;
//	cout<<(int)tree[index[0]].type;
//	cout.flush();
//	print();
//	j = this->depth(0);
	delete index;
	delete used;
	delete t;
}

int DecisionTree::depth(int pos)
{
	if (!tree[pos].used)
		log->error(logArea,"depth(...)");
	if (tree[pos].type != PREDICATE)
		return 1;
	else
		return 1+maxInt(depth(tree[pos].next1),depth(tree[pos].next2));
}

void DecisionTree::reduce()
{
	int i, pos=0;
	while (tree[pos].type!=ACTION)
		pos=randBool()?tree[pos].next1:tree[pos].next2;
	tree[0]=tree[pos];
	tree[0].level=1;
	for (i=1; i<maxEntries; ++i)
		tree[i].used=false;
}

void DecisionTree::_test()
{
	int t;
	for (int i=0; i<100; ++i) {
		t = rand();
		//cout<<i<<' '<<t<<'-'<<(long)0x7FFFFFFF<<endl; cout.flush();
		//if (t<0 || t>=0x8FFF)
		//	log->error(logArea,"_test()");
		if (t<0 || t>=0x80000000)
		  log->error(logArea,"_test() - long");
	}
}

void DecisionTree::mutate()
{
	int p = getPos(randInt(1,nNodes()));
	/*int d = tree[p].level+depth(p)-1;
	int d_min = tree[p].level;
	int d_max = MAX_TREE_SIZE;
	int _min = randInt(d_min,d);
	int _max = randInt(_min,minInt(d+2,d_max));
	*/
	int d = depth(0);
	d = minInt(d,tree[p].level);
	//log->debug(1,"min ",minInt(d,MAX_TREE_SIZE));
	deleteSubTree(p);
	//log->debug(1,"max ",MAX_TREE_SIZE);
	createRandomSubTree(d,minInt(d,MAX_TREE_SIZE),MAX_TREE_SIZE,p,UNKNOWN);
}
