// gen1.cpp : Defines the entry point for the console application.
//

//#include "stdafx.h"
#include "DecisionTree1.h"
#include "generation.h"
#include "genTypes.h"
#include <stdlib.h>
#include <iostream.h>
#include <string.h>

void genName(char* s, int g, bool stats, bool txt) {
  char str[] = "gen0000_.dat"; 	
  str[6] = '0'+(char)(g%10);
  str[5] = '0'+(g/10)%10;
  str[4] = '0'+(g/100)%10;
  str[3] = '0'+(g/1000)%10;
  if (txt) {
    str[9] ='t';
    str[10]='x';
    str[11]='t';
  }
  if (stats) {
    str[7]=str[8];
    str[8]=str[9];
    str[9]=str[10];
    str[10]=str[11];
    str[11]=0;
  }
  str[12]=0;
  int i=0;
  while (str[i]) {
    s[i]=str[i];
    i++;
  }
  s[i]=0;
}

int main(int argc, char* argv[])
{
  Log *log = new Log(true);

  int i;
	
  enum Script { New, Evolve, Test, Unknown } script;

  char filename[30];

  if (argc!=2 || !strcmp(argv[1],"-help")) {
    cout<<"Valid flags\n  -help\n  -new\n  -evolve -test\n";
    cout.flush();
    return 0;
  }
  else if (!strcmp(argv[1],"-new")) {
    script=New;
  }
  else if (!strcmp(argv[1],"-evolve")) {
    script=Evolve;
  }
  else if (!strcmp(argv[1],"-test")) {
    script=Test;
  }
  else
    script=Unknown;
  
  if (script == New) {
    Generation *gen = new Generation(log,0);
    gen->createNewPopulation(2,5,2,5);
    gen->save("gen.dat");
    genName(filename,0,false,false);
    gen->save(filename);
    genName(filename,0,false,true);
    gen->print(filename);
    delete gen;
  }
  else if (script == Evolve) {
    Generation *g1, *g2;
    g1 = new Generation(log,0);
    g1->load("gen.dat");
    g1->loadStats(STATFILE);
    genName(filename,g1->generation,true,false);
    g1->save(filename);
    genName(filename,g1->generation,true,true);
    g1->print(filename);
    g2 = g1->evolve();
    g2->save("gen.dat");
    genName(filename,g2->generation,false,false);
    g2->save(filename);
    genName(filename,g2->generation,false,true);
    g2->print(filename);
    delete g1;
    delete g2;
  }
  else if (script == Test) {
    Generation *g, *gtmp;
    cout<<"Testing select...";
    int p[POPULATION_SIZE];
    int n=1000000, k;
    g = new Generation(log,0);
    for (i=0; i<POPULATION_SIZE; ++i) {
      p[i]=0;
      g->inv_rank[i+1]=i;
    }
    for (i=0; i<n; ++i) {
      k = g->select();
      if (k<0 || k>=POPULATION_SIZE)
	cerr<<"Index Out Of Bound\n";
      p[k]++;
    }
    cout.precision(2);
    float q = (float)POPULATION_SIZE/(float)n;
    cout<<" N(0)="<<q*p[0]
	<<" N("<<POPULATION_SIZE/4<<")="<<q*p[POPULATION_SIZE/4]
	<<" N("<<POPULATION_SIZE/2<<")="<<q*p[POPULATION_SIZE/2]
	<<" N("<<3*POPULATION_SIZE/4<<")="<<q*p[3*POPULATION_SIZE/4]
	<<" N("<<POPULATION_SIZE-1<<")="<<q*p[POPULATION_SIZE-1]
	<<endl;
    cout.flush();
    delete g;
    cout<<"Testing evolving...\n";
    g = new Generation(log,0);
    g->createNewPopulation(2,5,2,5);
    for (i=0; i<1000; ++i) {
      g->loadStats(STATFILE);
      gtmp = g->evolve();
      delete g;
      g = gtmp;
    }
    genName(filename,g->generation,false,true);
    g->print(filename);
    delete g;
  }
  else 
    cout << "Unknown Script";


  delete log;
  return 0;
}
