// genTypes.h
//
//////////////////////////////////////////////////////////////////////

#if !defined(GENTYPES_H)
#define GENTYPES_H

//#if _MSC_VER > 1000
//#pragma once
//#endif // _MSC_VER > 1000

const int POPULATION_SIZE = 128;
const float QUOTA = 20.0;
const int MAX_TREE_SIZE = 7; //refers to the maximum depth of the tree and NOT the number of nodes/points
const char LOGFILE[] = "logevolve.txt";
const char STATFILE[]= "fitness.dat";
const float PC = 0.6; //Probability of crossover
const float PM = 0.2; //Probability of mutation

typedef enum TreeType {KICK, MOVE};

typedef enum NodeType {UNKNOWN, PREDICATE, ACTION };

typedef enum Predicates { 
  IAmClosestToBall,
  IAmSurrounded,
  _StartCommonPredicates,    //S Common
  WeAreWining,               //*
  IAmNearGoal,               //*
  IAmNearOurGoal,            //*
  OpponentIsClose,           //*
  IAmAlone,                  //*
  WeAreAttacking,
  OpponentsAreAttacking,
  TeammatesAreFree,
  WeAreSpreadOut,
  _EndCommonPredicates,      //Common
  _StartKickPredicates,      //S Kick
  _EndKickPredicates,        //Kick
  _StartMovePredicates,      //S Move
  BallIsClose,               //*
  BallIsFarAway,             //*
  BallIsNearGoal,            //*
  BallIsNearOurGoal,         //*
  IAmClosestToBallOnOurTeam, //*
  WeAreClosestToBall,        //*
  _EndMovePredicates         //Move
};

const char predicatesStrings[][30] = {
//   |<-0     |         |         |<-29
  "IAmClosestToBall",
  "IAmSurrounded",
  "_StartCommonPredicates",
  "WeAreWining",
  "IAmNearGoal",
  "IAmNearOurGoal",
  "OpponentIsClose",
  "IAmAlone",
  "WeAreAttacking",
  "OpponentsAreAttacking",
  "TeammatesAreFree",
  "WeAreSpreadOut",
  "_EndCommonPredicates",
  "_StartKickPredicates",
  "_EndKickPredicates",
  "_StartMovePredicates",
  "BallIsClose",
  "BallIsFarAway",
  "BallIsNearGoal",
  "BallIsNearOurGoal",
  "IAmClosestToBallOnOurTeam",
  "WeAreClosestToBall",
  "_EndMovePredicates"
};

typedef enum Actions { 
  _StartCommonActions,
  _EndCommonActions,
  _StartKickActions,
  PassPlayer1,
  PassPlayer2,
  PassPlayerFarAway,
  PassBeckham,
  DribbleToGoal,
  KickGoal,
  _EndKickActions,
  _StartMoveActions,
  DashToBall,
  DashToGoal,
  BlockGoal,
  InterceptBall,
  SpreadTeam,
  _EndMoveActions
};

const char actionStrings[][30] = {
//   |<-0     |         |         |<-29
  "_StartCommonActions",
  "_EndCommonActions",
  "_StartKickActions",
  "PassPlayer1",
  "PassPlayer2",
  "PassPlayerFarAway",
  "PassBeckham",
  "DribbleToGoal",
  "KickGoal",
  "_EndKickActions",
  "_StartMoveActions",
  "DashToBall",
  "DashToGoal",
  "BlockGoal",
  "InterceptBall",
  "SpreadTeam",
  "_EndMoveActions"

};

typedef union _nodeID {
	Actions action;
	Predicates predicate;
}NodeID;

typedef struct _dTree {
	// if "used" then this node is in the tree
	bool used;
	// this node's depth. The level of the root node is 1
	int level;
	// "type" defines the of node 
	NodeType type;
	// "id" defines the type of action/predicate associated with this node
	NodeID id;
	// "next1" and "next2" points to this node's children (note: next2 is only used if PREDICATE node) 
	int next1;
	int next2;
}dTree;

const int goals_size = 30;
typedef struct _Fitness {
  short inv;
  short ourScore;
  short theirScore;
  short scoringShoots;
  short kicks;
  short passes;
  short goals[goals_size];
} Fitness;

typedef struct _Tmp_Point {
  bool isKnown;
  float x;
  float y;
} Tmp_Point;

typedef struct _Position {
  short cycle;
  bool playon;
  Tmp_Point agentPos;
  Tmp_Point ballPos;
} Position;

#endif // GENTYPES_H
