#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <math.h>
#include "fitness.h"
#include "/home/jonatan/robocup/v2/evolve/genTypes.h"
#include "/home/jonatan/robocup/v2/scriptcreator/scconst.h"

//Constants
const char h[]={'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};

//Macros
#define MIN(x,y) ( (x)>(y)?(y):(x) )
#define HEX(x) h[x/16]<<h[x%16]
#define HEX1(x) h[x%16]
#define ERROR(x) cout<<"ERROR: "<<x<<endl;
#define IFERROR(y,x) if(y) ERROR(x)
#define PRINT(f,n,x) { char *s=itoa(x); for(int _k=1;_k<=MIN(n-strlen(s),9);++_k) f<<' '; f<<s; delete s; }

//Methods
char* itoa(int k) {
  char *s = new char[10];
  int i,j=0;
  bool n=(k<0);
  if (n)
    k=-k;
  s[0]=0;
  do {
    for (i=j; i>=0; --i)
      s[i+1]=s[i];
    s[0]='0'+k%10;
    k/=10;
    j++;
  }
  while (k>0);
  if (n) {
    for (i=j; i>=0; --i)
      s[i+1]=s[i];
    s[0]='-';
  }
  return s;
}


int maxIndex(int *array, int len) {
  int index=0, max=array[0]; 
  int i;
  for (i=1; i<len; ++i) 
    if (array[i]>max) {
      max=array[i];
      index=i;
    } 
  return index;
}

float distance(Tmp_Point p1, Tmp_Point p2, bool mirror) {
  double d;
  if (mirror)
    d=(p1.x+p2.x)*(p1.x+p2.x) + (p1.y+p2.y)*(p1.y+p2.y);
  else
    d=(p1.x-p2.x)*(p1.x-p2.x) + (p1.y-p2.y)*(p1.y-p2.y);
  return (float)sqrt(d);
}

void loadMatchData() { 
  int i,j,k,l; 
  FitData data[2*teamSize];
  Tmp_Point raw_ball[2][teamSize][3001];  // (2*10*3000)*9 = 500 kb
  Tmp_Point agent[2][teamSize][3001];
  Tmp_Point ball[3001];
  int on[3001]; //=1 if play on, 0 if other play mode, -1 if not set
  // init
  for (j=0; j<=1; ++j)
    for (k=0; k<=3000; ++k) {
      on[k]=-1;
      ball[k].isKnown=false;
      for (i=0; i<teamSize; ++i) {
	raw_ball[j][i][k].isKnown=false;
	agent[j][i][k].isKnown=false;
      }
    }
  // Load data from *.pos files
  for (j=0; j<=1; ++j)
    for (i=0; i<teamSize; ++i) {
      ifstream f;
      Position raw;
      if (j==0)
	f.open(posfilel[i],ios::in|ios::binary);
      else
	f.open(posfiler[i],ios::in|ios::binary);
      f.read((char *)&raw,sizeof(Position));
      while (!f.eof() && f.is_open()) {
	l = raw.cycle;
	if (l>=0 && l<=3000) {
	  if (!raw.playon || on[l]==0)
	    on[l]=0;
	  else
	    on[l]=1;
	  agent[j][i][l]=raw.agentPos;
	  raw_ball[j][i][l]=raw.ballPos;
	}
	else {
	  ERROR("input invalid (l="<<l<<")");
	}
	f.read((char *)&raw,sizeof(Position));
      }
    }
  // Debug
  for (k=0; k<=3000; ++k) 
    if (on[k]==1 && agent[0][0][k].isKnown) {
      cout<<agent[0][0][k].x<<' '<<agent[0][0][k].y<<endl;
    }
  // Determine the ball's average position for each cycle
  for (k=0; k<=3000; k++)
    if (on[k]==1) {
      ball[k].x=0.0;
      ball[k].y=0.0;
      int n=0;
      for (j=0; j<=1; j++)
	for (i=0; i<teamSize; i++) 
	  if (raw_ball[j][i][k].isKnown) {
	    n++;
	    if (j==0) { // left team player
	      ball[k].x+=raw_ball[j][i][k].x;
	      ball[k].y+=raw_ball[j][i][k].y;
	    }
	    else { // right team player
	      ball[k].x-=raw_ball[j][i][k].x;
	      ball[k].y-=raw_ball[j][i][k].y;
	    }
	  }
      if (n>0) {
	ball[k].isKnown=true;
	ball[k].x/=(float)n;
	ball[k].y/=(float)n;
      }
      else {
	ball[k].isKnown=false;
      }
    }
  // Determine avg_dis_to_ball for each player
  for (j=0; j<=1; j++)
    for (i=0; i<teamSize; i++) {
      float d=0;
      int n=0;
      for (k=0; k<=3000; ++k)
	if (on[k]==1 && ball[k].isKnown && agent[j][i][k].isKnown) {
	  n++;
	  d+=distance(ball[k],agent[j][i][k],j==1);
	}
      if (n>0) 
	d/=(float)n;
      else {
	ERROR("Unable to determine avg_dis_to_ball for agent (j,i)=("<<j<<','<<i<<")\n");
      }
#define INDEX j*teamSize+i
      data[INDEX].avg_dis_to_ball=(short)round(d);
    }
  // Determine avg_y & offensiv_time for each player
  for (j=0; j<=1; j++)
    for (i=0; i<teamSize; i++) {
      float d=0;
      int off=0;
      int n=0;
      for (k=0; k<=3000; ++k)
	if (on[k]==1 && agent[j][i][k].isKnown) {
	  n++;
	  d+=agent[j][i][k].y;
	  if (agent[j][i][k].y>0.0)
	    off++;
	}
      if (n>0) {
	cout<<off<<' '<<n<<endl;
	d/=(float)n;
	off*=100;
	off/=n;
      }
      else {
	ERROR("Unable to determine avg_y & offensive_time for agent (j,i)=("<<j<<','<<i<<")\n");
      }
      data[INDEX].avg_y=(short)round(d);
      data[INDEX].offensive_time=(short)off;
    }
  // Determine per_cent_open for each player
  for (j=0; j<=1; j++)
    for (i=0; i<teamSize; i++) {
      int n=0;
      int open=0;
      int ii, jj;
      for (k=0; k<=3000; ++k)
	if (on[k]==1 && agent[j][i][k].isKnown) {
	  bool ok=false;
	  float min=1000.0;
	  for (jj=0; jj<=1; jj++)
	    for (ii=0; ii<teamSize; ii++) 
	      if ( (jj!=j || ii!=i) && agent[jj][ii][k].isKnown) {
		ok=true;
		min=MIN(min,distance(agent[j][i][k],agent[jj][ii][k],j!=jj));
	      }
	  if (ok) {
	    n++;
	    if (min>=OPEN_DISTANCE) {
	      open++;
	    }
	  }
	}
      //cout<<open<<' '<<n<<endl;
      if (n>0) {
	open*=100;
	open/=n;
      }
      else {
	ERROR("Unable to determine per_cent_open for agent (j,i)=("<<j<<','<<i<<")\n");
      }
      data[INDEX].per_cent_open=(short)open;
    }
  // Load data from *.fit files
  for (j=0; j<=1; ++j) {
    Fitness raw;
    int goals[goals_size][teamSize]; 
    for (i=0; i<teamSize; ++i) {
      ifstream f;
      // open file
      if (j==0)
	f.open(fitfilel[i],ios::in|ios::binary);
      else
	f.open(fitfiler[i],ios::in|ios::binary);
      // reset data
      char *p = (char *)&raw;
      for (k=0; k<sizeof(Fitness); k++)
	p[k]=0;
      f.read((char *)&raw,sizeof(Fitness));
      // check if file existed
      bool eof = f.eof() || !f.is_open();
      // close file
      f.close(); 
      // set all data but positioning, myScore
      if (!eof) { 
	data[INDEX].inv           = raw.inv;
	data[INDEX].myScore       = 0;
	data[INDEX].ourScore      = raw.ourScore;
	data[INDEX].theirScore    = raw.theirScore;
	data[INDEX].kicks         = raw.kicks;
	data[INDEX].passes        = raw.passes;
	data[INDEX].scoringShoots = raw.scoringShoots;
	IFERROR(i>0 && l!=raw.ourScore,"fit"<<j<<i+1);
	l = raw.ourScore;
      }
      else
	ERROR("file does not exists - fit"<<j<<i+1);
      // save goal history
      for (k=0; k<goals_size; ++k)
	goals[k][i]=raw.goals[k];
    } 
    // determine scorer
    for (k=0; k<l; ++k) {
      i = maxIndex(goals[k],teamSize);
      data[INDEX].myScore++;
    }
  }
#undef INDEX
  // Write Match Log
  ofstream mf;
  mf.open(MATCHLOG,ios::out|ios::app);
  mf<<"----------------------------------------\n  Match ";
  for (i=0; i<teamSize; ++i)
    mf<<HEX(data[i].inv);
  mf<<" VS. ";
  for (i=teamSize; i<2*teamSize; ++i)
    mf<<HEX(data[i].inv);
  mf<<"  ("<<data[0].ourScore<<" - "<<data[0].theirScore<<")\n";
  mf<<"----------------------------------------\n";
  mf<<"    inv  goals  attemps  kicks  passes  ball_dis  avg_y  %open  %off.  \n";
  for (i=0; i<2*teamSize; ++i) {
    if (i==teamSize)
      mf<<'\n';
    mf<<((i>=teamSize)?'r':'l')<<HEX1(i%teamSize+1)<<": ";
    PRINT(mf,3,data[i].inv);
    PRINT(mf,7,data[i].myScore);
    PRINT(mf,9,data[i].scoringShoots);
    PRINT(mf,7,data[i].kicks);
    PRINT(mf,8,data[i].passes);
    PRINT(mf,10,data[i].avg_dis_to_ball);
    PRINT(mf,7,data[i].avg_y);
    PRINT(mf,7,data[i].per_cent_open);
    PRINT(mf,7,data[i].offensive_time);
    mf<<'\n';
  }
  mf<<"\n\n";
  mf.close(); 
  // Save Data
  ofstream df;
  df.open(INTERFILE,ios::out|ios::binary|ios::app);
  for (i=0; i<2*teamSize; ++i)
    df.write((char *)&data[i],sizeof(FitData));
  df.close(); 
}

void calc() {
  int i,j,k;
  // Load data from intermediate data file
  ifstream f;
  f.open(INTERFILE,ios::in|ios::binary);
  f.seekg(0,ios::end);
  int len_data = f.tellg()/sizeof(FitData);
  f.seekg(0,ios::beg);
  FitData *data = new FitData[len_data];
  for (int i=0; i<len_data; ++i)
    f.read((char *)&data[i],sizeof(FitData));
  f.close();
  // Set population size
  const int len_pop = POPULATION_SIZE;
  // Determine fitness
  float fit[len_pop];
  int cfit[len_pop];
  for (i=0; i<len_pop; ++i) {
    fit[i]=0.0;
    cfit[i]=0;
  }
  for (i=0; i<len_data; ++i)
    cfit[ data[i].inv ]++;
  for (i=0; i<len_data; ++i) 
    fit[ data[i].inv ]+= w.calc(data[i])/(float)cfit[ data[i].inv ];
  // Save fitness
  ofstream sf;
  sf.open(STATFILE,ios::out|ios::trunc|ios::binary);
  for (i=0; i<len_pop; ++i)
    sf.write((char *)&fit[i],sizeof(float));
  sf.close();
  // Print Stats over the Population
  ofstream pf, rf;
  pf.open(POPSTATS,ios::out|ios::trunc);
  rf.open(RAWPOPSTATS,ios::out|ios::trunc);
  pf<<"#Evaluations = "<<len_data<<" (should be = "<<len_pop*nOfGamesPerPlayer<<")\n\n";
  for (i=0; i<len_pop; ++i) {
    k=0;
    pf<<"\n------------------------------------------------------\n";
    pf<<"  Inv. # "<<i+1<<"  (fitness = "<<fit[i]<<")\n";
    pf<<"------------------------------------------------------\n";
    pf<<"     score  goals  attemps  kicks  passes  ball_dis  avg_y  %open  %off. game_fitness \n";
    for (j=0; j<len_data; ++j) {
      if (data[j].inv==i) {
	k++;
	PRINT(pf,3,k);
	pf<<". ";
	PRINT(pf,2,data[j].ourScore);
	pf<<"-";
	PRINT(pf,2,data[j].theirScore);
	PRINT(pf,7,data[j].myScore);
	PRINT(pf,9,data[j].scoringShoots);
	PRINT(pf,7,data[j].kicks);
	PRINT(pf,8,data[j].passes);
	PRINT(pf,10,data[j].avg_dis_to_ball);
	PRINT(pf,7,data[j].avg_y);
	PRINT(pf,7,data[j].per_cent_open);
	PRINT(pf,7,data[j].offensive_time);
	pf<<"    "<<w.calc(data[j])<<endl;
	// print to rf (for MATLAB)
	rf<<i<<' '<<k<<' '<<data[j].ourScore<<' '<<data[j].theirScore<<' '<<data[j].myScore<<' ';
	rf<<data[j].scoringShoots<<' '<<data[j].kicks<<' '<<data[j].passes<<' '<<data[j].avg_dis_to_ball<<' ';
	rf<<data[j].avg_y<<' '<<data[j].per_cent_open<<' '<<data[j].offensive_time<<' ';
	rf<<w.calc(data[j])<<endl;
      }
      else if (data[j].inv==len_pop)
	pf<<"\nERROR\n";
    }
  }
  pf.close();
  rf.close();
  delete data;
}

void reset() {
  ofstream f;
  f.open(INTERFILE,ios::out|ios::binary|ios::trunc);
  f.close();
}

void help() {
  cout<<"\nValid options:\n";
  cout<<"  -reset    Reset the intermediate file\n";
  cout<<"  -load     Load data from last match\n";
  cout<<"  -calc     Determine fitness for the whole population\n\n";
}

//Main
int main(int argc, char** argv) {
  if (argc>1)
    if (strcmp(argv[1],"-reset")==0)
      reset();
    else if (strcmp(argv[1],"-load")==0)
      loadMatchData();
    else if (strcmp(argv[1],"-calc")==0)
      calc();
    else
      help();
  else
    help();
  return 1;
}

//#undef HEX
//#undef HEX1
//#undef PRINT
//#undef IFERROR
