// agent's body dir is supposed to be known
bool TeamDecision::alignBody1(AngleDeg ad) {
  AngleDegUE anue = agentView->getNeckDirection();
  if (anue.isUnknown())
    return false;
  AngleDeg an = anue.getAngle()-ad;
  if (ABS(an)<5.0)
    return false;
  DLLOG(" ALIGN! "<<an);
  ServerCommand* cmd;
  cmd = new TurnCommand(an);
  actuatorInterface->addToPlan(cmd);
  delete cmd;
  cmd = new TurnNeckCommand(-an);
  actuatorInterface->addToPlan(cmd);
  delete cmd;
  return true;
}

int _cFreeKick = 0;
bool TeamDecision::decide_freekick() {
  cout<<"...\n"; cout.flush();
  if ((ballView->getRelativeDistance() <= facts->SP_kickable_margin)) {
    ServerCommand *cmd;
    AngleDeg ang = (agentView->getPosition().getPoint()).getDirectionTo(facts->getTheirGoal().getPosition())-agentView->getBodyDirection().getAngle();
    if (ABS(ang)>10) {
      cout<<"1\n"; cout.flush();
      cmd = new TurnCommand(ang);
      //_cFreeKick=0;
    }
    else if (_cFreeKick<2) {
      cout<<"2\n"; cout.flush();
      cmd = new TurnCommand(0);
      _cFreeKick++;
    }
    else {
      cout<<"3\n"; cout.flush();
      _cFreeKick=0;
      cmd = new KickCommand(100,ang);
    }
    /*
    PlayerObject po = playerView->getTeammateClosestToAgent();
    if (po.getPosition().isKnown()) {
      AngleDegUE po_dir = agentView->getPosition().getDirectionTo(po.getPosition());
      scmd = new KickCommand(50,(po_dir-agentView->getBodyDirection()).getAngle());
    }
    else {
      scmd = new TurnCommand(40);
    }
    */
    actuatorInterface->addToPlan(cmd);
    delete cmd;
  }
  else
    decideNoAction();
}

void TeamDecision::onAuralRefereeSensorData() {
  PlayMode mode = gameView->getPlayMode();
  if (mode == PM_OUR_GOAL || mode == PM_THEIR_GOAL)
      initialPositioning(0);
}

void TeamDecision::initialPositioning(int state) {
  Point mp = Point(_X, 0);
  //int gtime = (int)(gameView->getCurrentTime()).getT();
  AngleDegUE ad = agentView->getBodyDirection();
  ServerCommand* cmd;
  if (state == 0) {
    cmd = new MoveCommand(mp);
  }
  else if (state == 1 && ad.isKnown()) {
    AngleDeg ang = dirToCenter-ad.getAngle();
    if (abs(ang)<5) 
      ang = 0.0;
    cmd = new TurnCommand(ang);
  }
  else
    cmd = new TurnCommand(0);
  actuatorInterface->addToPlan(cmd);
  delete cmd;
}

void TeamDecision::onInit() {
  dirToCenter = Point(-10,0).getDirectionTo(Point(0,0));
  initialPositioning(0);
  side = (gameView->getOurSide()==S_LEFT)?'l':'r';
}

//int lgtime=-1;
ServerCommand *persistentCmd = NULL;
void TeamDecision::onCommandWarning() {
  PlayMode mode = gameView->getPlayMode();
  
  PointUE bpUE = ballView->getPosition();
  TimeUnknown ballobsU = ballView->getLastObservationTime();
  //int gtime = (int)(gameView->getCurrentTime()).getT();
  IFLLOG(gtime == lgtime,"MORE THAN ONE WARNINGS IN SAME CYCLE","")
  IFLLOG(gtime > lgtime+1,"MISSED CYCLE","")
    //lgtime = gtime;
  if (bpUE.isKnown()) {
    ballIsKnown = true;
    current_bp = bpUE.getPoint();
    //if (ballobsU.isKnown())
    //  ballobs = gtime-ballobsU.getT();
    //else
    //  ballobs = -10000;
  }
  else
    ballIsKnown = false;
  /*
    PLAY ON
  */ 
  if (mode == PM_PLAY_ON) { 
    if (persistentCmd != NULL) {
      LLOG("Persistent CMD");
      actuatorInterface->addToPlan(persistentCmd);
      delete persistentCmd;
      persistentCmd = NULL;
      return;
    }
    PointUE ap = agentView->getPosition();
    AngleDegUE ad = agentView->getBodyDirection();
    float balldis;
    if (ap.isUnknown() || ad.isUnknown()) {
      LLOG("AGENT UNKNOWN");
      decideNoAction();
      return;
    }
    else if (ballIsKnown)
      balldis = ballView->getRelativeDistance().getValue(); 
    
    
    if (!ballIsKnown || 
	//(ballobs>=10 && ballobs%5 && balldis>40) ||
	(balldis>40)) {
      IFLLOG(!ballIsKnown,"BALL UNKNOWN","")
      ELSE
      IFLLOG((ballobs>=10 && ballobs%5),"WHERE IS THE BALL","BALL FAR AWAY")
      ServerCommand *cmd;
      if (alignBody1(ad.getAngle()))
	return;
      if ((ap.getPoint()).getDistanceTo(Point(_X,0))>2.0) {
	LLOG("GOTO HOME dir="<<(ap.getPoint()).getDirectionTo(Point(_X,0))<<" ad="<<ad.getAngle())
	AngleDeg ang = (ap.getPoint()).getDirectionTo(Point(_X,0))-ad.getAngle();
	if (ABS(ang)>5.0) { 
	  cmd = new TurnCommand(ang); 
	  persistentCmd = new DashCommand(50);
	}
	else
	  cmd = new DashCommand(50);
      }
      else {
	LLOG("SCANING FROM HOME POS")
	AngleDeg ang = dirToCenter-ad.getAngle();
	if (ABS(ang)>5.0) {
	  cmd = new TurnCommand(dirToCenter-ad.getAngle());
	  persistentCmd = new TurnCommand(0);
	}
	else if (ang<0) //gtime%4==0 || (gtime+1)%4==0)
	  cmd = new TurnCommand(45);
	else
	  cmd = new TurnCommand(-45);
      }
      actuatorInterface->addToPlan(cmd);
      delete cmd;
    }
    else { // ball is known and ball is closer than 40 meters
      if ( balldis<10) {
	LLOG("BALLISCLOSE")
	if (alignBody1(ad.getAngle()))
	  return;
#define PENAREA(x) ( x.getX()<-32 && x.getY()>-20 && x.getY()<20 )
	if (catchBall.applicable().isTrue() && PENAREA(ap.getPoint())) {
	  LLOG("CATCH BALL IS APPLICABLE");
	  static Plan plan = catchBall.generatePlan();
	  actuatorInterface->addToPlan(plan.getNextStep()); 
	}
#undef PENAREA
	else if (balldis<=facts->SP_kickable_area) {
	  ServerCommand *cmd;
	  cmd = new KickCommand(70,dirToCenter-ad.getAngle());
	  delete cmd;
	}
	else {
	  LLOG("GOTO BALL")
	    //static Plan plan = gotoBall.generatePlan();
	    //actuatorInterface->addToPlan(plan.getNextStep());
	  Point bp;
	  if (ballView->getPosition(NEXT).isKnown())
	    bp = ballView->getPosition(NEXT).getPoint();
	  else
	    bp = current_bp;
	  LLOG("BP = "<<bp)
	  LLOG("ap.dirTo(bp)"<<(ap.getPoint()).getDirectionTo(bp))
	  LLOG("ad = "<<ad.getAngle())
	  AngleDeg ang = (ap.getPoint()).getDirectionTo(bp)-ad.getAngle();
	  LLOG("ang = "<<ang)
	  ServerCommand *cmd;
	  if (ABS(ang)>5.0) {
	    cmd = new TurnCommand(ang);
	    persistentCmd = new DashCommand(100);
	  }
	  else
	    cmd = new DashCommand(100);
	  actuatorInterface->addToPlan(cmd);
	  delete cmd;
	}
      }
      /*
	else if (balldis<20) {
	if (alignBody(ad.getAngle()))
	return;
	ServerCommand *cmd;
	AngleDeg ang = ballView->getRelativeDirection()-ad.getAngle();
	if (ang>5.0 || ang<-5.0) {
	VectorUE sv = ballView->getSpeedVector();
	if (sv.isKnown()) {
	    
	}
	else
	cmd = new TurnCommand(0);
	}
	else
	cmd = new TurnCommand(ang);
	actuatorInterface->addToPlan(cmd);
	delete cmd;
	}
      */
      else {
	LLOG("Entering Move-On-Line-Skill");
	// determine move point
	Point target = current_bp;
	target.setX(_X);
	if (target.getY() > 10)
	  target.setY(10);
	else if (target.getY() < -10)
	  target.setY(-10);
       	// determine neck angle
	AngleDeg neckturn = ballView->getRelativeDirection().getAngle()-agentView->getNeckDirection().getAngle();
	// determine dash/turn command
	AngleDeg ang = (ap.getPoint()).getDirectionTo(target)-ad.getAngle();
	ServerCommand *cmd;
	if ((ap.getPoint()).getDistanceTo(target)>1.0) {
	  if (ABS(ang)>5.0) {
	    cmd = new TurnCommand(ang);
	    neckturn -= ang;
	  }
	  else
	    cmd = new DashCommand(50);
	}
	else
	  cmd = new TurnCommand(0);
	LLOG("MOVE ON LINE, neck="<<neckturn<<", body="<<ang);
	// turn neck & move to target
      	if (ABS(neckturn)>5.0) {
	  ServerCommand *ncmd;
	  LLOG("TURNING NECK");
	  ncmd = new TurnNeckCommand(neckturn);
	  actuatorInterface->addToPlan(ncmd);
	  delete ncmd;
	}
	actuatorInterface->addToPlan(cmd);
	delete cmd;
      }
	

      //static Plan plan = goaliePositioning.generatePlan();
      //actuatorInterface->addToPlan(plan.getNextStep()); 
    

    }
     

    /*
      decide();
    
  
      PointUE bp = ballView->getPosition();
      PointUE ap = agentView->getPosition();
      LLOG((int)(gameView->getCurrentTime()).getT())
      if (findBall.applicable()) {
      LLOG("FINDBALL")
      static Plan plan = findBall.generatePlan();
      actuatorInterface->addToPlan(plan.getNextStep());
      }
      else if (bp.isKnown() && ap.isKnown()) {
      if (ap.getDistanceTo(bp)>40) {
      LLOG("MOVE HOME")
      moveToPosition.setPoint(Point(-46,0));
      static Plan plan = moveToPosition.generatePlan();
      actuatorInterface->addToPlan(plan.getNextStep());
      } 
      else {
      LLOG("POS")
      static Plan plan = goaliePositioning.generatePlan();
      actuatorInterface->addToPlan(plan.getNextStep());
      }
      } 
      else 
      decideNoAction();
    */
  }
  /*
    OTHER PLAY MODES
  */
  else { 
    persistentCmd = NULL;
    if (mode == PM_OUR_KICK_OFF || mode == PM_THEIR_KICK_OFF) {
      LLOG("KICK OFF")
      initialPositioning(1);
    }
    else if (mode == PM_OUR_GOAL_KICK) {
      LLOG("GOAL KICK");
      if (ballView->getRelativeDistance() <= facts->SP_kickable_margin) {
	if (agentView->getPosition().isUnknown() ||
	    agentView->getBodyDirection().isUnknown() ||
	    !ballIsKnown) {
	  decideNoAction();
	  return;
	}
	decide_freekick();
      }
      else if (findBall.applicable()) {
	static Plan plan = findBall.generatePlan();
	actuatorInterface->addToPlan(plan.getNextStep());
      }
      else if (ballIsKnown) {
	if (agentView->getPosition().isUnknown() ||
	    agentView->getBodyDirection().isUnknown()) {
	  decideNoAction();
	  return;
	}
	Point ap = agentView->getPosition().getPoint();
	AngleDeg ad = agentView->getBodyDirection().getAngle();
	AngleDeg ang = ap.getDirectionTo(current_bp)-ad;
	ServerCommand *cmd;
	if (ABS(ang)>5.0)
	  cmd = new TurnCommand(ang);
	else
	  cmd = new DashCommand(60);
	actuatorInterface->addToPlan(cmd);
	delete cmd;
      }
      else
	decideNoAction();
    }
    else if (mode == PM_OUR_FREE_KICK) {
      LLOG("FREE KICK");
      if (agentView->getPosition().isUnknown() ||
	  agentView->getBodyDirection().isUnknown()) {
	decideNoAction();
      }
      else
	decide_freekick();
    }
    else {
      LLOG("UNKNOWN PLAY MODE");
      decideNoAction();
    } 
  }
}

#undef LLOG
#undef DLLOG
#undef IFLLOG
#undef ELSE
#undef _X
#undef ABS

RS_END_NAMESPACE
