/*
 *  udp UDPSocket program.
 */
#include "udpsocket.h"



UDPSocket::UDPSocket( void ) {
	zerotime.tv_sec = 0;
	zerotime.tv_usec = 0;
	//timePassed( 0, 1 );
}



UDPSocket::~UDPSocket( void ) {
	close( sock.socketfd );
}


 
/***************************************************************
 * Function sets up the socket, connects to host and prepares
 * a few variables for function pollSocket.
 * Returns -1 in case of failure
 */
int UDPSocket::initConnection( char *host, int port ) {
	struct hostent *host_ent;
	struct in_addr	*addr_ptr;
	struct sockaddr_in	cli_addr;
	int sockfd, val;

	sock.socketfd = -1;

	if( ( host_ent = (struct hostent *)gethostbyname( host ) ) == NULL )
	/* Check if a numeric address */
	if( inet_addr( host ) == -1 )
			return sock.socketfd;

	addr_ptr = (struct in_addr *) *host_ent->h_addr_list;
	host = inet_ntoa(*addr_ptr);


	/*
	 *  Open UDP socket.
	 */
	if( ( sockfd = socket( AF_INET, SOCK_DGRAM, 0 ) ) < 0 )
		return sock.socketfd;	/* Can't open socket. */

	val = fcntl( sockfd, F_GETFL, 0 );
#if defined NewsOS || defined IRIX
	val |= O_NDELAY;
#else
	val |= O_NONBLOCK;
#endif
	val |= O_NDELAY;
	
	fcntl( sockfd, F_SETFL, val );		
	
	/*
	 *  Bind any local address.
	 */
	//bzero( (char *) &cli_addr, sizeof( cli_addr ) );
	memset( (char *) &cli_addr,0,sizeof( cli_addr ) );
	cli_addr.sin_family = AF_INET;
	cli_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	cli_addr.sin_port = htons(0);

	if( bind( sockfd, (struct sockaddr *) &cli_addr, sizeof( cli_addr ) ) >= 0 ) {
		/*
		 *  Fill in the structure with the address of the server.
		 */
		sock.socketfd = sockfd;
	
		//bzero( (char *) &sock.serv_addr, sizeof( sock.serv_addr ) );
		memset((char *) &sock.serv_addr,0,sizeof( sock.serv_addr ) );
		sock.serv_addr.sin_family = AF_INET;
		sock.serv_addr.sin_addr.s_addr = inet_addr( host );
		sock.serv_addr.sin_port = htons( port );
	}
	
	/*
	 * Prepare for pollSocket
	 */
	FD_ZERO( &readfds );
	FD_SET( sock.socketfd, &readfds );
	readfds_bak = readfds;
	
	return sock.socketfd;
}



/***************************************************************
 * Function sends message located in buf via socket. Returns -1
 * in case of failure
 */
int UDPSocket::sendMessage( char *buf ) {
	int n;

	n = strlen(buf);
	//printf( "send %d : %s\n", ntohs( sock.serv_addr.sin_port ), buf );
	if( sendto( sock.socketfd, buf, n, 0, (struct sockaddr *)&sock.serv_addr, sizeof( sock.serv_addr ) ) != n )
		return -1;
	
	return 0;
}


/***************************************************************
 * Function receives message waiting on socket placing the
 * characters in buf. Returns number of characters read
 * or -1 if failure
 */
int UDPSocket::receiveMessage( char *buf ) {
	int n;
	struct sockaddr_in serv_addr;
	int servlen;
		
	servlen = sizeof( serv_addr );
	n = recvfrom( sock.socketfd, buf, MAXMSG, 0, (struct sockaddr *)&serv_addr, &(socklen_t)servlen );

	if( n < 0 ) {
		if( n == -1 && errno == EWOULDBLOCK ) {
			buf[0] = '\0';
			return 0;
		}		
		
		return -1;
	}

	//printf( "recv %d : ", ntohs( serv_addr.sin_port ) );
	sock.serv_addr.sin_port = serv_addr.sin_port;
	buf[n] = '\0';

	if( !n )
		return 0;
		
	return 1;
}



/***************************************************************
 * Function polls socket and returns true if there is a message
 * waiting.
 */
bool UDPSocket::pollSocket( void ) {
	readfds = readfds_bak;

	if( select( sock.socketfd + 1, &readfds, NULL, NULL, &zerotime ) < 0 ) {
		perror( "Error in UDPSocket::pollSocket select()\n" );
		exit( 1 );
	}
	
	if( FD_ISSET( sock.socketfd, &readfds ) )
		return true;

	return false;
}


void UDPSocket::closeConnection() {
	close( sock.socketfd );
}
